from __future__ import unicode_literals
from .base import Completion, Completer, ThreadedCompleter, DummyCompleter, DynamicCompleter, CompleteEvent, merge_completers, get_common_complete_suffix
from .filesystem import PathCompleter, ExecutableCompleter
from .word_completer import WordCompleter
from .fuzzy_completer import FuzzyCompleter, FuzzyWordCompleter

__all__ = [
    # Base.
    'Completion',
    'Completer',
    'ThreadedCompleter',
    'DummyCompleter',
    'DynamicCompleter',
    'CompleteEvent',
    'merge_completers',
    'get_common_complete_suffix',

    # Filesystem.
    'PathCompleter',
    'ExecutableCompleter',

    # Word completer.
    'WordCompleter',
    'FuzzyCompleter',
    'FuzzyWordCompleter',
]



"""
The default styling.
"""
from __future__ import unicode_literals, absolute_import
from .style import Style, merge_styles
from .base import ANSI_COLOR_NAMES
from .named_colors import NAMED_COLORS
from prompt_toolkit.cache import memoized

__all__ = [
    'default_ui_style',
    'default_pygments_style',
]

#: Default styling. Mapping from classnames to their style definition.
PROMPT_TOOLKIT_STYLE = [
    # Highlighting of search matches in document.
    ('search',                             'bg:ansibrightyellow ansiblack'),
    ('search.current',                     ''),

    # Incremental search.
    ('incsearch',                          ''),
    ('incsearch.current',                  'reverse'),

    # Highlighting of select text in document.
    ('selected',                                'reverse'),

    ('cursor-column',                           'bg:#dddddd'),
    ('cursor-line',                             'underline'),
    ('color-column',                            'bg:#ccaacc'),

    # Highlighting of matching brackets.
    ('matching-bracket',                        ''),
    ('matching-bracket.other',                  '#000000 bg:#aacccc'),
    ('matching-bracket.cursor',                 '#ff8888 bg:#880000'),

    # Styling of other cursors, in case of block editing.
    ('multiple-cursors',                        '#000000 bg:#ccccaa'),

    # Line numbers.
    ('line-number',                             '#888888'),
    ('line-number.current',                     'bold'),
    ('tilde',                                   '#8888ff'),

    # Default prompt.
    ('prompt',                                  ''),
    ('prompt.arg',                              'noinherit'),
    ('prompt.arg.text',                         ''),
    ('prompt.search',                           'noinherit'),
    ('prompt.search.text',                      ''),

    # Search toolbar.
    ('search-toolbar',                          'bold'),
    ('search-toolbar.text',                     'nobold'),

    # System toolbar
    ('system-toolbar',                          'bold'),
    ('system-toolbar.text',                     'nobold'),

    # "arg" toolbar.
    ('arg-toolbar',                             'bold'),
    ('arg-toolbar.text',                        'nobold'),

    # Validation toolbar.
    ('validation-toolbar',                      'bg:#550000 #ffffff'),
    ('window-too-small',                        'bg:#550000 #ffffff'),

    # Completions toolbar.
    ('completion-toolbar',                     'bg:#bbbbbb #000000'),
    ('completion-toolbar.arrow',               'bg:#bbbbbb #000000 bold'),
    ('completion-toolbar.completion',          'bg:#bbbbbb #000000'),
    ('completion-toolbar.completion.current',  'bg:#444444 #ffffff'),

    # Completions menu.
    ('completion-menu',                         'bg:#bbbbbb #000000'),
    ('completion-menu.completion',              ''),
    ('completion-menu.completion.current',      'bg:#888888 #ffffff'),
    ('completion-menu.meta.completion',         'bg:#999999 #000000'),
    ('completion-menu.meta.completion.current', 'bg:#aaaaaa #000000'),
    ('completion-menu.multi-column-meta',       'bg:#aaaaaa #000000'),

    # Fuzzy matches in completion menu (for FuzzyCompleter).
    ('completion-menu.completion fuzzymatch.outside',          'fg:#444444'),
    ('completion-menu.completion fuzzymatch.inside',           'bold'),
    ('completion-menu.completion fuzzymatch.inside.character', 'underline'),
    ('completion-menu.completion.current fuzzymatch.outside',  'fg:default'),
    ('completion-menu.completion.current fuzzymatch.inside',   'nobold'),

    # Styling of readline-like completions.
    ('readline-like-completions',                                        ''),
    ('readline-like-completions.completion',                             ''),
    ('readline-like-completions.completion fuzzymatch.outside',          '#888888'),
    ('readline-like-completions.completion fuzzymatch.inside',           ''),
    ('readline-like-completions.completion fuzzymatch.inside.character', 'underline'),

    # Scrollbars.
    ('scrollbar.background',                     'bg:#aaaaaa'),
    ('scrollbar.button',                         'bg:#444444'),
    ('scrollbar.arrow',                          'noinherit bold'),

    # Start/end of scrollbars. Adding 'underline' here provides a nice little
    # detail to the progress bar, but it doesn't look good on all terminals.
    # ('scrollbar.start',                          'underline #ffffff'),
    # ('scrollbar.end',                            'underline #000000'),

    # Auto suggestion text.
    ('auto-suggestion',                         '#666666'),

    # Trailing whitespace and tabs.
    ('trailing-whitespace',                     '#999999'),
    ('tab',                                     '#999999'),

    # When Control-C/D has been pressed. Grayed.
    ('aborting',  '#888888 bg:default noreverse noitalic nounderline noblink'),
    ('exiting',   '#888888 bg:default noreverse noitalic nounderline noblink'),

    # Entering a Vi digraph.
    ('digraph',                                 '#4444ff'),

    # Control characters, like ^C, ^X.
    ('control-character',                       'ansiblue'),

    # Non-breaking space.
    ('nbsp',                                    'underline ansiyellow'),

    # Default styling of HTML elements.
    ('i',                                       'italic'),
    ('u',                                       'underline'),
    ('b',                                       'bold'),
    ('em',                                      'italic'),
    ('strong',                                  'bold'),
    ('hidden',                                  'hidden'),

    # It should be possible to use the style names in HTML.
    # <reverse>...</reverse>  or <noreverse>...</noreverse>.
    ('italic',                                  'italic'),
    ('underline',                               'underline'),
    ('bold',                                    'bold'),
    ('reverse',                                 'reverse'),
    ('noitalic',                                'noitalic'),
    ('nounderline',                             'nounderline'),
    ('nobold',                                  'nobold'),
    ('noreverse',                               'noreverse'),

    # Prompt bottom toolbar
    ('bottom-toolbar',                          'reverse'),
]


# Style that will turn for instance the class 'red' into 'red'.
COLORS_STYLE = [
    (name, 'fg:' + name) for name in ANSI_COLOR_NAMES
] + [
    (name.lower(), 'fg:' + name) for name in NAMED_COLORS
]


WIDGETS_STYLE = [
    # Dialog windows.
    ('dialog',                                  'bg:#4444ff'),
    ('dialog.body',                             'bg:#ffffff #000000'),
    ('dialog.body text-area',                    'bg:#cccccc'),
    ('dialog.body text-area last-line',          'underline'),

    ('dialog frame.label',                      '#ff0000 bold'),

    # Scrollbars in dialogs.
    ('dialog.body scrollbar.background',        ''),
    ('dialog.body scrollbar.button',            'bg:#000000'),
    ('dialog.body scrollbar.arrow',             ''),
    ('dialog.body scrollbar.start',             'nounderline'),
    ('dialog.body scrollbar.end',               'nounderline'),

    # Buttons.
    ('button',                                  ''),
    ('button.arrow',                            'bold'),
    ('button.focused',                          'bg:#aa0000 #ffffff'),

    # Menu bars.
    ('menu-bar',                                'bg:#aaaaaa #000000'),
    ('menu-bar.selected-item',                  'bg:#ffffff #000000'),
    ('menu',                                    'bg:#888888 #ffffff'),
    ('menu.border',                             '#aaaaaa'),
    ('menu.border shadow',                      '#444444'),

    # Shadows.
    ('dialog shadow',                           'bg:#000088'),
    ('dialog.body shadow',                      'bg:#aaaaaa'),

    ('progress-bar',                             'bg:#000088'),
    ('progress-bar.used',                        'bg:#ff0000'),
]


# The default Pygments style, include this by default in case a Pygments lexer
# is used.
PYGMENTS_DEFAULT_STYLE = {
    'pygments.whitespace':                "#bbbbbb",
    'pygments.comment':                   "italic #408080",
    'pygments.comment.preproc':           "noitalic #bc7a00",

    'pygments.keyword':                   "bold #008000",
    'pygments.keyword.pseudo':            "nobold",
    'pygments.keyword.type':              "nobold #b00040",

    'pygments.operator':                  "#666666",
    'pygments.operator.word':             "bold #aa22ff",

    'pygments.name.builtin':              "#008000",
    'pygments.name.function':             "#0000ff",
    'pygments.name.class':                "bold #0000ff",
    'pygments.name.namespace':            "bold #0000ff",
    'pygments.name.exception':            "bold #d2413a",
    'pygments.name.variable':             "#19177c",
    'pygments.name.constant':             "#880000",
    'pygments.name.label':                "#a0a000",
    'pygments.name.entity':               "bold #999999",
    'pygments.name.attribute':            "#7d9029",
    'pygments.name.tag':                  "bold #008000",
    'pygments.name.decorator':            "#aa22ff",

    # Note: In Pygments, Token.String is an alias for Token.Literal.String,
    #       and Token.Number as an alias for Token.Literal.Number.
    'pygments.literal.string':            "#ba2121",
    'pygments.literal.string.doc':        "italic",
    'pygments.literal.string.interpol':   "bold #bb6688",
    'pygments.literal.string.escape':     "bold #bb6622",
    'pygments.literal.string.regex':      "#bb6688",
    'pygments.literal.string.symbol':     "#19177c",
    'pygments.literal.string.other':      "#008000",
    'pygments.literal.number':            "#666666",

    'pygments.generic.heading':           "bold #000080",
    'pygments.generic.subheading':        "bold #800080",
    'pygments.generic.deleted':           "#a00000",
    'pygments.generic.inserted':          "#00a000",
    'pygments.generic.error':             "#ff0000",
    'pygments.generic.emph':              "italic",
    'pygments.generic.strong':            "bold",
    'pygments.generic.prompt':            "bold #000080",
    'pygments.generic.output':            "#888",
    'pygments.generic.traceback':         "#04d",

    'pygments.error':                     "border:#ff0000",
}


@memoized()
def default_ui_style():
    """
    Create a default `Style` object.
    """
    return merge_styles([
        Style(PROMPT_TOOLKIT_STYLE),
        Style(COLORS_STYLE),
        Style(WIDGETS_STYLE),
    ])


@memoized()
def default_pygments_style():
    """
    Create a `Style` object that contains the default Pygments style.
    """
    return Style.from_dict(PYGMENTS_DEFAULT_STYLE)



"""
Search related key bindings.
"""
from __future__ import unicode_literals
from ..key_bindings import key_binding
from prompt_toolkit.application.current import get_app
from prompt_toolkit.filters import is_searching, control_is_searchable, Condition
from prompt_toolkit import search

__all__ = [
    'abort_search',
    'accept_search',
    'start_reverse_incremental_search',
    'start_forward_incremental_search',
    'reverse_incremental_search',
    'forward_incremental_search',
    'accept_search_and_accept_input',
]


@key_binding(filter=is_searching)
def abort_search(event):
    """
    Abort an incremental search and restore the original
    line.
    (Usually bound to ControlG/ControlC.)
    """
    search.stop_search()


@key_binding(filter=is_searching)
def accept_search(event):
    """
    When enter pressed in isearch, quit isearch mode. (Multiline
    isearch would be too complicated.)
    (Usually bound to Enter.)
    """
    search.accept_search()


@key_binding(filter=control_is_searchable)
def start_reverse_incremental_search(event):
    """
    Enter reverse incremental search.
    (Usually ControlR.)
    """
    search.start_search(direction=search.SearchDirection.BACKWARD)


@key_binding(filter=control_is_searchable)
def start_forward_incremental_search(event):
    """
    Enter forward incremental search.
    (Usually ControlS.)
    """
    search.start_search(direction=search.SearchDirection.FORWARD)


@key_binding(filter=is_searching)
def reverse_incremental_search(event):
    """
    Apply reverse incremental search, but keep search buffer focused.
    """
    search.do_incremental_search(
        search.SearchDirection.BACKWARD, count=event.arg)


@key_binding(filter=is_searching)
def forward_incremental_search(event):
    """
    Apply forward incremental search, but keep search buffer focused.
    """
    search.do_incremental_search(
        search.SearchDirection.FORWARD, count=event.arg)


@Condition
def _previous_buffer_is_returnable():
    """
    True if the previously focused buffer has a return handler.
    """
    prev_control = get_app().layout.search_target_buffer_control
    return prev_control and prev_control.buffer.is_returnable


@key_binding(filter=is_searching & _previous_buffer_is_returnable)
def accept_search_and_accept_input(event):
    """
    Accept the search operation first, then accept the input.
    """
    search.accept_search()
    event.current_buffer.validate_and_handle()



from __future__ import unicode_literals
import inspect
import os
import signal
import sys
import threading
import weakref

from collections import deque
from functools import partial
from six import PY2, text_type
from six.moves import range
from wcwidth import wcwidth
from .cache import memoized

__all__ = [
    'Event',
    'DummyContext',
    'get_cwidth',
    'suspend_to_background_supported',
    'is_conemu_ansi',
    'is_windows',
    'in_main_thread',
    'take_using_weights',
    'test_callable_args',
    'to_str',
    'to_int',
    'to_float',
]


class Event(object):
    """
    Simple event to which event handlers can be attached. For instance::

        class Cls:
            def __init__(self):
                # Define event. The first parameter is the sender.
                self.event = Event(self)

        obj = Cls()

        def handler(sender):
            pass

        # Add event handler by using the += operator.
        obj.event += handler

        # Fire event.
        obj.event()
    """
    def __init__(self, sender, handler=None):
        self.sender = sender
        self._handlers = []

        if handler is not None:
            self += handler

    def __call__(self):
        " Fire event. "
        for handler in self._handlers:
            handler(self.sender)

    def fire(self):
        " Alias for just calling the event. "
        self()

    def add_handler(self, handler):
        """
        Add another handler to this callback.
        (Handler should be a callable that takes exactly one parameter: the
        sender object.)
        """
        # Test handler.
        assert callable(handler)
        if not _func_takes_one_arg(handler):
            raise TypeError("%r doesn't take exactly one argument." % handler)

        # Add to list of event handlers.
        self._handlers.append(handler)

    def remove_handler(self, handler):
        """
        Remove a handler from this callback.
        """
        if handler in self._handlers:
            self._handlers.remove(handler)

    def __iadd__(self, handler):
        " `event += handler` notation for adding a handler. "
        self.add_handler(handler)
        return self

    def __isub__(self, handler):
        " `event -= handler` notation for removing a handler. "
        self.remove_handler(handler)
        return self


# Cache of signatures. Improves the performance of `test_callable_args`.
_signatures_cache = weakref.WeakKeyDictionary()

_inspect_signature = getattr(inspect, 'signature', None)  # Only on Python 3.


def test_callable_args(func, args):
    """
    Return True when this function can be called with the given arguments.
    """
    assert isinstance(args, (list, tuple))

    if _inspect_signature is not None:
        # For Python 3, use inspect.signature.
        try:
            sig = _signatures_cache[func]
        except KeyError:
            sig = _inspect_signature(func)
            _signatures_cache[func] = sig

        try:
            sig.bind(*args)
        except TypeError:
            return False
        else:
            return True
    else:
        # For older Python versions, fall back to using getargspec
        # and don't check for `partial`.
        if isinstance(func, partial):
            return True

        spec = inspect.getargspec(func)

        # Drop the 'self'
        def drop_self(spec):
            args, varargs, varkw, defaults = spec
            if args[0:1] == ['self']:
                args = args[1:]
            return inspect.ArgSpec(args, varargs, varkw, defaults)

        spec = drop_self(spec)

        # When taking *args, always return True.
        if spec.varargs is not None:
            return True

        # Test whether the given amount of args is between the min and max
        # accepted argument counts.
        return len(spec.args) - len(spec.defaults or []) <= len(args) <= len(spec.args)


@memoized(maxsize=1024)
def _func_takes_one_arg(func):
    """
    Test whether the given function can be called with exactly one argument.
    """
    return test_callable_args(func, [None])


class DummyContext(object):
    """
    (contextlib.nested is not available on Py3)
    """
    def __enter__(self):
        pass

    def __exit__(self, *a):
        pass


class _CharSizesCache(dict):
    """
    Cache for wcwidth sizes.
    """
    LONG_STRING_MIN_LEN = 64  # Minimum string length for considering it long.
    MAX_LONG_STRINGS = 16  # Maximum number of long strings to remember.

    def __init__(self):
        super(_CharSizesCache, self).__init__()
        # Keep track of the "long" strings in this cache.
        self._long_strings = deque()

    def __missing__(self, string):
        # Note: We use the `max(0, ...` because some non printable control
        #       characters, like e.g. Ctrl-underscore get a -1 wcwidth value.
        #       It can be possible that these characters end up in the input
        #       text.
        if len(string) == 1:
            result = max(0, wcwidth(string))
        else:
            result = sum(self[c] for c in string)

        # Store in cache.
        self[string] = result

        # Rotate long strings.
        # (It's hard to tell what we can consider short...)
        if len(string) > self.LONG_STRING_MIN_LEN:
            long_strings = self._long_strings
            long_strings.append(string)

            if len(long_strings) > self.MAX_LONG_STRINGS:
                key_to_remove = long_strings.popleft()
                if key_to_remove in self:
                    del self[key_to_remove]

        return result


_CHAR_SIZES_CACHE = _CharSizesCache()


def get_cwidth(string):
    """
    Return width of a string. Wrapper around ``wcwidth``.
    """
    return _CHAR_SIZES_CACHE[string]


def suspend_to_background_supported():
    """
    Returns `True` when the Python implementation supports
    suspend-to-background. This is typically `False' on Windows systems.
    """
    return hasattr(signal, 'SIGTSTP')


def is_windows():
    """
    True when we are using Windows.
    """
    return sys.platform.startswith('win')  # E.g. 'win32', not 'darwin' or 'linux2'


def is_windows_vt100_supported():
    """
    True when we are using Windows, but VT100 escape sequences are supported.
    """
    # Import needs to be inline. Windows libraries are not always available.
    from prompt_toolkit.output.windows10 import is_win_vt100_enabled
    return is_windows() and is_win_vt100_enabled()


def is_conemu_ansi():
    """
    True when the ConEmu Windows console is used.
    """
    return is_windows() and os.environ.get('ConEmuANSI', 'OFF') == 'ON'


def in_main_thread():
    """
    True when the current thread is the main thread.
    """
    return threading.current_thread().__class__.__name__ == '_MainThread'


def get_term_environment_variable():
    " Return the $TERM environment variable. "
    term = os.environ.get('TERM', '')
    if PY2:
        term = term.decode('utf-8')
    return term


def take_using_weights(items, weights):
    """
    Generator that keeps yielding items from the items list, in proportion to
    their weight. For instance::

        # Getting the first 70 items from this generator should have yielded 10
        # times A, 20 times B and 40 times C, all distributed equally..
        take_using_weights(['A', 'B', 'C'], [5, 10, 20])

    :param items: List of items to take from.
    :param weights: Integers representing the weight. (Numbers have to be
                    integers, not floats.)
    """
    assert isinstance(items, list)
    assert isinstance(weights, list)
    assert all(isinstance(i, int) for i in weights)
    assert len(items) == len(weights)
    assert len(items) > 0

    # Remove items with zero-weight.
    items2 = []
    weights2 = []
    for i, w in zip(items, weights):
        if w > 0:
            items2.append(i)
            weights2.append(w)

    items = items2
    weights = weights2

    # Make sure that we have some items left.
    if not items:
        raise ValueError("Did't got any items with a positive weight.")

    #
    already_taken = [0 for i in items]
    item_count = len(items)
    max_weight = max(weights)

    i = 0
    while True:
        # Each iteration of this loop, we fill up until by (total_weight/max_weight).
        adding = True
        while adding:
            adding = False

            for item_i, item, weight in zip(range(item_count), items, weights):
                if already_taken[item_i] < i * weight / float(max_weight):
                    yield item
                    already_taken[item_i] += 1
                    adding = True

        i += 1


def to_str(value):
    " Turn callable or string into string. "
    if callable(value):
        return to_str(value())
    else:
        return text_type(value)


def to_int(value):
    " Turn callable or int into int. "
    if callable(value):
        return to_int(value())
    else:
        return int(value)


def to_float(value):
    " Turn callable or float into float. "
    if callable(value):
        return to_float(value())
    else:
        return float(value)



from __future__ import unicode_literals
from prompt_toolkit.styles.pygments import pygments_token_to_classname
from .base import FormattedText

__all__ = [
    'PygmentsTokens',
]


class PygmentsTokens(object):
    """
    Turn a pygments token list into a list of prompt_toolkit text fragments
    (``(style_str, text)`` tuples).
    """
    def __init__(self, token_list):
        assert isinstance(token_list, list), 'Got %r' % (token_list, )
        self.token_list = token_list

    def __pt_formatted_text__(self):
        result = FormattedText()

        for token, text in self.token_list:
            result.append(('class:' + pygments_token_to_classname(token), text))

        return result



from __future__ import unicode_literals
import re

from collections import namedtuple
from six import string_types

from prompt_toolkit.document import Document
from prompt_toolkit.filters import to_filter

from .base import Completer, Completion
from .word_completer import WordCompleter

__all__ = [
    'FuzzyCompleter',
    'FuzzyWordCompleter',
]


class FuzzyCompleter(Completer):
    """
    Fuzzy completion.
    This wraps any other completer and turns it into a fuzzy completer.

    If the list of words is: ["leopard" , "gorilla", "dinosaur", "cat", "bee"]
    Then trying to complete "oar" would yield "leopard" and "dinosaur", but not
    the others, because they match the regular expression 'o.*a.*r'.
    Similar, in another application "djm" could expand to "django_migrations".

    The results are sorted by relevance, which is defined as the start position
    and the length of the match.

    Notice that this is not really a tool to work around spelling mistakes,
    like what would be possible with difflib. The purpose is rather to have a
    quicker or more intuitive way to filter the given completions, especially
    when many completions have a common prefix.

    Fuzzy algorithm is based on this post:
    https://blog.amjith.com/fuzzyfinder-in-10-lines-of-python

    :param completer: A :class:`~.Completer` instance.
    :param WORD: When True, use WORD characters.
    :param pattern: Regex pattern which selects the characters before the
        cursor that are considered for the fuzzy matching.
    :param enable_fuzzy: (bool or `Filter`) Enabled the fuzzy behavior. For
        easily turning fuzzyness on or off according to a certain condition.
    """
    def __init__(self, completer, WORD=False, pattern=None, enable_fuzzy=True):
        assert isinstance(completer, Completer)
        assert pattern is None or pattern.startswith('^')

        self.completer = completer
        self.pattern = pattern
        self.WORD = WORD
        self.pattern = pattern
        self.enable_fuzzy = to_filter(enable_fuzzy)

    def get_completions(self, document, complete_event):
        if self.enable_fuzzy():
            return self._get_fuzzy_completions(document, complete_event)
        else:
            return self.completer.get_completions(document, complete_event)

    def _get_pattern(self):
        if self.pattern:
            return self.pattern
        if self.WORD:
            return r'[^\s]+'
        return '^[a-zA-Z0-9_]*'

    def _get_fuzzy_completions(self, document, complete_event):
        word_before_cursor = document.get_word_before_cursor(
            pattern=re.compile(self._get_pattern()))

        # Get completions
        document2 = Document(
            text=document.text[:document.cursor_position - len(word_before_cursor)],
            cursor_position=document.cursor_position - len(word_before_cursor))

        completions = list(self.completer.get_completions(document2, complete_event))

        fuzzy_matches = []
        pat = '.*?'.join(map(re.escape, word_before_cursor))
        pat = '(?=({0}))'.format(pat)   # lookahead regex to manage overlapping matches
        regex = re.compile(pat, re.IGNORECASE)
        for compl in completions:
            matches = list(regex.finditer(compl.text))
            if matches:
                # Prefer the match, closest to the left, then shortest.
                best = min(matches, key=lambda m: (m.start(), len(m.group(1))))
                fuzzy_matches.append(_FuzzyMatch(len(best.group(1)), best.start(), compl))

        def sort_key(fuzzy_match):
            " Sort by start position, then by the length of the match. "
            return fuzzy_match.start_pos, fuzzy_match.match_length

        fuzzy_matches = sorted(fuzzy_matches, key=sort_key)

        for match in fuzzy_matches:
            # Include these completions, but set the correct `display`
            # attribute and `start_position`.
            yield Completion(
                match.completion.text,
                start_position=match.completion.start_position - len(word_before_cursor),
                display_meta=match.completion.display_meta,
                display=self._get_display(match, word_before_cursor),
                style=match.completion.style)

    def _get_display(self, fuzzy_match, word_before_cursor):
        """
        Generate formatted text for the display label.
        """
        m = fuzzy_match
        word = m.completion.text

        if m.match_length == 0:
            # No highlighting when we have zero length matches (no input text).
            return word

        result = []

        # Text before match.
        result.append(('class:fuzzymatch.outside', word[:m.start_pos]))

        # The match itself.
        characters = list(word_before_cursor)

        for c in word[m.start_pos:m.start_pos + m.match_length]:
            classname = 'class:fuzzymatch.inside'
            if characters and c.lower() == characters[0].lower():
                classname += '.character'
                del characters[0]

            result.append((classname, c))

        # Text after match.
        result.append(
            ('class:fuzzymatch.outside', word[m.start_pos + m.match_length:]))

        return result


class FuzzyWordCompleter(Completer):
    """
    Fuzzy completion on a list of words.

    (This is basically a `WordCompleter` wrapped in a `FuzzyCompleter`.)

    :param words: List of words or callable that returns a list of words.
    :param meta_dict: Optional dict mapping words to their meta-information.
    :param WORD: When True, use WORD characters.
    """
    def __init__(self, words, meta_dict=None, WORD=False):
        assert callable(words) or all(isinstance(w, string_types) for w in words)

        self.words = words
        self.meta_dict = meta_dict or {}
        self.WORD = WORD

        self.word_completer = WordCompleter(
            words=lambda: self.words,
            WORD=self.WORD)

        self.fuzzy_completer = FuzzyCompleter(
            self.word_completer,
            WORD=self.WORD)

    def get_completions(self, document, complete_event):
        return self.fuzzy_completer.get_completions(document, complete_event)


_FuzzyMatch = namedtuple('_FuzzyMatch', 'match_length start_pos completion')



from __future__ import unicode_literals
from .base import Input, DummyInput
from .defaults import create_input, get_default_input, set_default_input

__all__ = [
    # Base.
    'Input',
    'DummyInput',

    # Defaults.
    'create_input',
    'get_default_input',
    'set_default_input',
]



# pylint: disable=function-redefined
from __future__ import unicode_literals

from prompt_toolkit.application.current import get_app
from prompt_toolkit.filters import has_selection, Condition, emacs_insert_mode, vi_insert_mode, in_paste_mode, is_multiline
from prompt_toolkit.key_binding.key_processor import KeyPress
from prompt_toolkit.keys import Keys

from .named_commands import get_by_name
from ..key_bindings import KeyBindings

__all__ = [
    'load_basic_bindings',
]


def if_no_repeat(event):
    """ Callable that returns True when the previous event was delivered to
    another handler. """
    return not event.is_repeat


def load_basic_bindings():
    key_bindings = KeyBindings()
    insert_mode = vi_insert_mode | emacs_insert_mode
    handle = key_bindings.add

    @handle('c-a')
    @handle('c-b')
    @handle('c-c')
    @handle('c-d')
    @handle('c-e')
    @handle('c-f')
    @handle('c-g')
    @handle('c-h')
    @handle('c-i')
    @handle('c-j')
    @handle('c-k')
    @handle('c-l')
    @handle('c-m')
    @handle('c-n')
    @handle('c-o')
    @handle('c-p')
    @handle('c-q')
    @handle('c-r')
    @handle('c-s')
    @handle('c-t')
    @handle('c-u')
    @handle('c-v')
    @handle('c-w')
    @handle('c-x')
    @handle('c-y')
    @handle('c-z')
    @handle('f1')
    @handle('f2')
    @handle('f3')
    @handle('f4')
    @handle('f5')
    @handle('f6')
    @handle('f7')
    @handle('f8')
    @handle('f9')
    @handle('f10')
    @handle('f11')
    @handle('f12')
    @handle('f13')
    @handle('f14')
    @handle('f15')
    @handle('f16')
    @handle('f17')
    @handle('f18')
    @handle('f19')
    @handle('f20')
    @handle('c-@')  # Also c-space.
    @handle('c-\\')
    @handle('c-]')
    @handle('c-^')
    @handle('c-_')
    @handle('backspace')
    @handle('up')
    @handle('down')
    @handle('right')
    @handle('left')
    @handle('s-up')
    @handle('s-down')
    @handle('s-right')
    @handle('s-left')
    @handle('home')
    @handle('end')
    @handle('delete')
    @handle('s-delete')
    @handle('c-delete')
    @handle('pageup')
    @handle('pagedown')
    @handle('s-tab')
    @handle('tab')
    @handle('c-left')
    @handle('c-right')
    @handle('c-up')
    @handle('c-down')
    @handle('insert')
    @handle(Keys.Ignore)
    def _(event):
        """
        First, for any of these keys, Don't do anything by default. Also don't
        catch them in the 'Any' handler which will insert them as data.

        If people want to insert these characters as a literal, they can always
        do by doing a quoted insert. (ControlQ in emacs mode, ControlV in Vi
        mode.)
        """
        pass

    # Readline-style bindings.
    handle('home')(get_by_name('beginning-of-line'))
    handle('end')(get_by_name('end-of-line'))
    handle('left')(get_by_name('backward-char'))
    handle('right')(get_by_name('forward-char'))
    handle('c-up')(get_by_name('previous-history'))
    handle('c-down')(get_by_name('next-history'))
    handle('c-l')(get_by_name('clear-screen'))

    handle('c-k', filter=insert_mode)(get_by_name('kill-line'))
    handle('c-u', filter=insert_mode)(get_by_name('unix-line-discard'))
    handle('backspace', filter=insert_mode, save_before=if_no_repeat)(
        get_by_name('backward-delete-char'))
    handle('delete', filter=insert_mode, save_before=if_no_repeat)(
        get_by_name('delete-char'))
    handle('c-delete', filter=insert_mode, save_before=if_no_repeat)(
        get_by_name('delete-char'))
    handle(Keys.Any, filter=insert_mode, save_before=if_no_repeat)(
        get_by_name('self-insert'))
    handle('c-t', filter=insert_mode)(get_by_name('transpose-chars'))
    handle('c-i', filter=insert_mode)(get_by_name('menu-complete'))
    handle('s-tab', filter=insert_mode)(get_by_name('menu-complete-backward'))

    # Control-W should delete, using whitespace as separator, while M-Del
    # should delete using [^a-zA-Z0-9] as a boundary.
    handle('c-w', filter=insert_mode)(get_by_name('unix-word-rubout'))

    handle('pageup', filter= ~has_selection)(get_by_name('previous-history'))
    handle('pagedown', filter= ~has_selection)(get_by_name('next-history'))

    # CTRL keys.

    text_before_cursor = Condition(lambda: get_app().current_buffer.text)
    handle('c-d', filter=text_before_cursor & insert_mode)(get_by_name('delete-char'))

    @handle('enter', filter=insert_mode & is_multiline)
    def _(event):
        " Newline (in case of multiline input. "
        event.current_buffer.newline(copy_margin=not in_paste_mode())

    @handle('c-j')
    def _(event):
        r"""
        By default, handle \n as if it were a \r (enter).
        (It appears that some terminals send \n instead of \r when pressing
        enter. - at least the Linux subsystem for Windows.)
        """
        event.key_processor.feed(
            KeyPress(Keys.ControlM, '\r'), first=True)

    # Delete the word before the cursor.

    @handle('up')
    def _(event):
        event.current_buffer.auto_up(count=event.arg)

    @handle('down')
    def _(event):
        event.current_buffer.auto_down(count=event.arg)

    @handle('delete', filter=has_selection)
    def _(event):
        data = event.current_buffer.cut_selection()
        event.app.clipboard.set_data(data)

    # Global bindings.

    @handle('c-z')
    def _(event):
        """
        By default, control-Z should literally insert Ctrl-Z.
        (Ansi Ctrl-Z, code 26 in MSDOS means End-Of-File.
        In a Python REPL for instance, it's possible to type
        Control-Z followed by enter to quit.)

        When the system bindings are loaded and suspend-to-background is
        supported, that will override this binding.
        """
        event.current_buffer.insert_text(event.data)

    @handle(Keys.BracketedPaste)
    def _(event):
        " Pasting from clipboard. "
        data = event.data

        # Be sure to use \n as line ending.
        # Some terminals (Like iTerm2) seem to paste \r\n line endings in a
        # bracketed paste. See: https://github.com/ipython/ipython/issues/9737
        data = data.replace('\r\n', '\n')
        data = data.replace('\r', '\n')

        event.current_buffer.insert_text(data)

    @handle(Keys.Any, filter=Condition(lambda: get_app().quoted_insert), eager=True)
    def _(event):
        """
        Handle quoted insert.
        """
        event.current_buffer.insert_text(event.data, overwrite=False)
        event.app.quoted_insert = False

    return key_bindings



from __future__ import unicode_literals
from prompt_toolkit.eventloop.context import TaskLocal, TaskLocalNotSetError
from contextlib import contextmanager

__all__ = [
    'get_app',
    'set_app',
    'NoRunningApplicationError',
]


_current_app = TaskLocal()


def get_app(raise_exception=False, return_none=False):
    """
    Get the current active (running) Application.
    An :class:`.Application` is active during the
    :meth:`.Application.run_async` call.

    We assume that there can only be one :class:`.Application` active at the
    same time. There is only one terminal window, with only one stdin and
    stdout. This makes the code significantly easier than passing around the
    :class:`.Application` everywhere.

    If no :class:`.Application` is running, then return by default a
    :class:`.DummyApplication`. For practical reasons, we prefer to not raise
    an exception. This way, we don't have to check all over the place whether
    an actual `Application` was returned.

    (For applications like pymux where we can have more than one `Application`,
    we'll use a work-around to handle that.)

    :param raise_exception: When `True`, raise
        :class:`.NoRunningApplicationError` instead of returning a
        :class:`.DummyApplication` if no application is running.
    :param return_none: When `True`, return `None`
        instead of returning a :class:`.DummyApplication` if no application is
        running.
    """
    try:
        value = _current_app.get()
    except TaskLocalNotSetError:
        if raise_exception:
            raise NoRunningApplicationError
        elif return_none:
            return None
        else:
            from .dummy import DummyApplication
            return DummyApplication()
    else:
        return value


@contextmanager
def set_app(app):
    """
    Context manager that sets the given :class:`.Application` active.

    (Usually, not needed to call outside of prompt_toolkit.)
    """
    from .application import Application
    assert app is None or isinstance(app, Application)

    previous = get_app(return_none=True)
    _current_app.set(app)

    try:
        yield
    finally:
        if previous:
            _current_app.set(previous)
        else:
            _current_app.delete()


class NoRunningApplicationError(Exception):
    " There is no active application right now. "



"""
Key bindings for extra page navigation: bindings for up/down scrolling through
long pages, like in Emacs or Vi.
"""
from __future__ import unicode_literals
from prompt_toolkit.filters import emacs_mode, buffer_has_focus
from .scroll import scroll_forward, scroll_backward, scroll_half_page_up, scroll_half_page_down, scroll_one_line_up, scroll_one_line_down, scroll_page_up, scroll_page_down
from prompt_toolkit.filters import vi_mode
from prompt_toolkit.key_binding.key_bindings import KeyBindings, ConditionalKeyBindings, merge_key_bindings

__all__ = [
    'load_page_navigation_bindings',
    'load_emacs_page_navigation_bindings',
    'load_vi_page_navigation_bindings',
]


def load_page_navigation_bindings():
    """
    Load both the Vi and Emacs bindings for page navigation.
    """
    # Only enable when a `Buffer` is focused, otherwise, we would catch keys
    # when another widget is focused (like for instance `c-d` in a
    # ptterm.Terminal).
    return ConditionalKeyBindings(
        merge_key_bindings([
            load_emacs_page_navigation_bindings(),
            load_vi_page_navigation_bindings(),
        ]), buffer_has_focus)


def load_emacs_page_navigation_bindings():
    """
    Key bindings, for scrolling up and down through pages.
    This are separate bindings, because GNU readline doesn't have them.
    """
    key_bindings = KeyBindings()
    handle = key_bindings.add

    handle('c-v')(scroll_page_down)
    handle('pagedown')(scroll_page_down)
    handle('escape', 'v')(scroll_page_up)
    handle('pageup')(scroll_page_up)

    return ConditionalKeyBindings(key_bindings, emacs_mode)


def load_vi_page_navigation_bindings():
    """
    Key bindings, for scrolling up and down through pages.
    This are separate bindings, because GNU readline doesn't have them.
    """
    key_bindings = KeyBindings()
    handle = key_bindings.add

    handle('c-f')(scroll_forward)
    handle('c-b')(scroll_backward)
    handle('c-d')(scroll_half_page_down)
    handle('c-u')(scroll_half_page_up)
    handle('c-e')(scroll_one_line_down)
    handle('c-y')(scroll_one_line_up)
    handle('pagedown')(scroll_page_down)
    handle('pageup')(scroll_page_up)

    return ConditionalKeyBindings(key_bindings, vi_mode)



"""
Future implementation for the prompt_toolkit eventloop.
"""
from __future__ import unicode_literals, print_function
from .base import EventLoop
from .context import get_context_id, context
from .defaults import get_event_loop
import sys

__all__ = [
    'Future',
    'InvalidStateError',
]


class InvalidStateError(Exception):
    " The operation is not allowed in this state. "


class Future(object):
    """
    `Future` object for use with the prompt_toolkit event loops.  (Not by
    accident very similar to asyncio -- but much more limited in functionality.
    They are however not meant to be used interchangeable.)
    """
    def __init__(self, loop=None):
        assert loop is None or isinstance(loop, EventLoop)
        self.loop = loop or get_event_loop()
        self.done_callbacks = []
        self._result = None
        self._exception = None
        self._done = False
        self._retrieved_result = False

        # Keep track of which `TaskContext` was active when this Future was
        # created.  This is the one that will be viewed as visible when the
        # callbacks are called.  (This is required to make get_app/set_app work
        # together with coroutines, when there are multiple active
        # applications, attached to different outputs.)
        self._ctx_id = get_context_id()

    # Thanks to asyncio for the following destructor!
    # On Python 3.3 and older, objects with a destructor part of a reference
    # cycle are never destroyed. It's not more the case on Python 3.4 thanks
    # to the PEP 442.
    if sys.version_info >= (3, 4):
        def __del__(self):
            if self._exception and not self._retrieved_result:
                exc = self._exception
                context = {
                    'message': ('%s exception was never retrieved'
                                % self.__class__.__name__),
                    'exception': exc,
                    'future': self,
                }
                self.loop.call_exception_handler(context)

    @classmethod
    def succeed(cls, result):
        """
        Returns a Future for which the result has been set to the given result.
        Similar to Twisted's `Deferred.succeed()`.
        """
        f = cls()
        f.set_result(result)
        return f

    @classmethod
    def fail(cls, result):
        """
        Returns a Future for which the error has been set to the given result.
        Similar to Twisted's `Deferred.fail()`.
        """
        f = cls()
        f.set_exception(result)
        return f

    def add_done_callback(self, callback):
        """
        Add a callback to be run when the future becomes done.  (This
        callback will be called with one argument only: this future
        object.)
        """
        self.done_callbacks.append(callback)

        # When a result has already been set. Call callback right away.
        if self._done:
            def call_cb():
                self._retrieved_result = True
                callback(self)

            self.loop.call_from_executor(call_cb)

    def set_result(self, result):
        " Mark the future done and set its result. "
        if self._done:
            raise InvalidStateError('Future result has been set already.')

        self._result = result
        self._done = True
        self._call_callbacks()

    def set_exception(self, exception):
        " Mark the future done and set an exception. "
        if self._done:
            raise InvalidStateError('Future result has been set already.')

        self._exception = exception
        self._done = True

        if self.done_callbacks:
            self._call_callbacks()
        else:
            # When an exception is set on a 'Future' object, but there
            # is no callback set to handle it, print the exception.
            # -- Uncomment for debugging. --

            # import traceback, sys
            # print(''.join(traceback.format_stack()), file=sys.__stderr__)
            # print('Uncollected error: %r' % (exception, ), file=sys.__stderr__)
            pass

    def _call_callbacks(self):
        # Create a local copy of the callbacks. Otherwise, it could be that
        # another call to `add_done_callback` would add a new callback to this list
        # which would then be called twice. (Once from here, once from the
        # `add_done_callback` function directly.
        done_callbacks = self.done_callbacks[:]

        if done_callbacks:
            self._retrieved_result = True

        def call_them_all():
            # Activate the original task context (and application) again.
            with context(self._ctx_id):
                # They should be called in order.
                for cb in done_callbacks:
                    cb(self)

        self.loop.call_from_executor(call_them_all)

    def result(self):
        " Return the result this future represents. "
        if not self._done:
            raise InvalidStateError

        self._retrieved_result = True
        if self._exception:
            raise self._exception
        else:
            return self._result

    def exception(self):
        " Return the exception that was set on this future. "
        if not self._done:
            raise InvalidStateError

        self._retrieved_result = True
        return self._exception

    def done(self):
        """
        Return True if the future is done. Done means either that a result /
        exception are available, or that the future was cancelled.
        """
        return self._done

    def to_asyncio_future(self):
        """
        Turn this `Future` into an asyncio `Future` object.
        """
        from asyncio import Future
        asyncio_f = Future()

        @self.add_done_callback
        def _(f):
            if f.exception():
                asyncio_f.set_exception(f.exception())
            else:
                asyncio_f.set_result(f.result())

        return asyncio_f

    @classmethod
    def from_asyncio_future(cls, asyncio_f, loop=None):
        """
        Return a prompt_toolkit `Future` from the given asyncio Future.
        """
        f = cls(loop=loop)

        @asyncio_f.add_done_callback
        def _(asyncio_f):
            if asyncio_f.exception():
                f.set_exception(asyncio_f.exception())
            else:
                f.set_result(asyncio_f.result())

        return f

    def __iter__(self):
        " For compatibility with asyncio. "
        return self.to_asyncio_future().__iter__()

    __await__ = __iter__
